#!/usr/bin/env python
#
#       scenegraph.py
#       
#       Copyright 2009 Gordon Allott <gordallott@gmail.com>
#       
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

from collections import deque 
#deque is important, saves on a crap load of memory


""" This is our scenegraph setup, we are using a quadtree approach
    For speed/simplicity reasons i am not using a rectangle based quadtree, 
    rather everything is point based, This can potentially mean the loss of
    precision which is not great, but understandable... just make sure that
    you call hit() on a larger area than you want, area_size + size of biggest item
    is a good idea
"""

class Rect(object):

    def __init__(self, x, y, width, height):
        self.x = float(x)
        self.y = float(y)
        self.width = float(width)
        self.height = float(height)
        
        self.left = self.x - (self.width * 0.5)
        self.right = self.x + (self.width * 0.5)
        self.top = self.y + (self.height * 0.5)
        self.bottom = self.y - (self.height * 0.5)
        
    def split(self):
        new_w = self.width * 0.5
        new_h = self.height * 0.5
        
        modx = new_w * 0.5
        mody = new_h * 0.5
        
        nw = Rect(self.x - modx, self.y + mody, new_w, new_h)
        ne = Rect(self.x + modx, self.y + mody, new_w, new_h)
        sw = Rect(self.x - modx, self.y - mody, new_w, new_h)
        se = Rect(self.x + modx, self.y - mody, new_w, new_h)
        
        return nw, ne, sw, se
        
        
    def point_in_bounds(self, point):
        """ checks to make sure point given is inside this rect """
        x, y = point
        if x >= self.left and x <= self.right and y <= self.top and y >= self.bottom:
            return True
        
        return False
        
    def rect_in_bounds(self, bounds):
        """ 
        checks to see if *any* part of the bounds provided fall inside this rect 
        """
        if self.left > bounds.right or bounds.left > self.right or \
        self.top < bounds.bottom or bounds.top < self.bottom:
            return False
        return True
        
    def __str__(self):
        return "Rect object: at %f -> %f --- %f -> %f" % (self.left, self.right, self.top, self.bottom) 

class Node(object):
    
    max_items = 8
    min_size = 0.1
    
    def __init__(self, bounds=None, parent=None):
        # we *always* start as a leaf node, we split into further buckets if we need to
        self.is_leaf = True
        self.items = deque()
        self.bounds = bounds
        self.parent = parent
        
    def add_item(self, item):
        """ make sure we can get at the item.position """
    
        if self.is_leaf:
            #we want the item in here
            self.items.append(item)
            item.qnode = self
            
            if self.bounds.width <= self.min_size:
                # if we are smaller than the min_size, we don't want to split anymore
                # problems with floating point precision cause infinate reccursion =\
                return
                
            
            if len(self.items) > self.max_items:
                #we breached capacity, we want to split our node into a bucket 
                # with four leafs

                #create new bounding boxes
                bnw, bne, bsw, bse = self.bounds.split()
                
                self.nw = Node(bounds=bnw, parent=self)
                self.ne = Node(bounds=bne, parent=self)
                self.sw = Node(bounds=bsw, parent=self)
                self.se = Node(bounds=bse, parent=self)
                
                self.is_leaf = False
                
                # we now need to add our items to the right leaf node for its
                # position
                
                for newitem in self.items:
                    if self.nw.bounds.point_in_bounds(newitem.position):
                        self.nw.add_item(newitem)
                        
                    elif self.ne.bounds.point_in_bounds(newitem.position):
                        self.ne.add_item(newitem)
                        
                    elif self.sw.bounds.point_in_bounds(newitem.position):
                        self.sw.add_item(newitem)
                        
                    elif self.se.bounds.point_in_bounds(newitem.position):
                        self.se.add_item(newitem)
                    else:
                        print self.bounds
                        print newitem
                        raise Exception, "Oh-oh, this can't happen - couldn't find place for %s" % (newitem)
                        
        
        else:
            # if we get here then we are a bucket, buckets just pass data down 
            # to the appropriate leaf node
            if self.nw.bounds.point_in_bounds(item.position):
                self.nw.add_item(item)
                
            elif self.ne.bounds.point_in_bounds(item.position):
                self.ne.add_item(item)
                
            elif self.sw.bounds.point_in_bounds(item.position):
                self.sw.add_item(item)
                
            elif self.se.bounds.point_in_bounds(item.position):
                self.se.add_item(item)
            else:
                print self.bounds
                print item
                raise Exception, "Oh-oh, this can't happen, couldn't find a place for %s" % (item)
            
                
    def hit(self, bounds):
        """ give this a set of bounds and it will return all the items that are 
        within those bounds, it may also return items outside it
        """
        if not self.is_leaf:
            # if we are a node and not a leaf we want to gather the results from
            # our quadrants not ourself
            hits = set()
            
            if bounds.rect_in_bounds(self.nw.bounds):
                hits |= self.nw.hit(bounds)
            
            if bounds.rect_in_bounds(self.ne.bounds):
                hits |= self.ne.hit(bounds)
            
            if bounds.rect_in_bounds(self.sw.bounds):
                hits |= self.sw.hit(bounds)
            
            if bounds.rect_in_bounds(self.se.bounds):
                hits |= self.se.hit(bounds)
            
            return hits
            
        else:
            # if we get here then we must be a leaf node, leaf nodes just need to
            # empty their contents
            return set(self.items)
    
    def remove_item(self, item):
        print [item]
        print self.items
        self.items.remove(item)
        
    def move_item(self, item):
        """ this will move an item from where it is now in the tree to where it
        *should* be depending on its position, we use this instead of just removing
        it and adding it again because we can navigate the tree backwards to
        reduce the size of the path we go on.
        
        I think this way the worst case scenario is much worse but the best case
        is much *much* better :) - just don't move things too quickly ;)
        """
        if self.is_leaf:
            if self.bounds.point_in_bounds(item.position):
                return
                
            #the item no longer fits in this leaf node, ask the parent to
            #figure out where it goes
            self.items.remove(item)
            item.qnode = None
            self.parent.move_item(item)
            
        else:
            #we are a node not a leaf
            if self.bounds.point_in_bounds(item.position):
                # our item is within our bounds, try adding it here
                self.add_item(item)
                
            else:
                self.parent.move_item(item)
